import pytest
from PilLite import Image


def test_create():
    img = Image.new(80, 60, 'RGB', 0xff0000)
    assert img.size == (80, 60)


@pytest.fixture(params=[
    dict(fmt='L', color=0xff),
    dict(fmt='L', color=0xa0),
    dict(fmt='L', color=0x0a),
    dict(fmt='LA', color=0xff00),
    dict(fmt='LA', color=0x00ff),
    dict(fmt='RGB', color=0xff0000),
    dict(fmt='RGB', color=0x00ff00),
    dict(fmt='RGB', color=0x0000ff),
    dict(fmt='RGBA', color=0x0000ff00),
    dict(fmt='RGBA', color=0x0000ffff),
    dict(fmt='RGBA', color=0xff00ff00),
    dict(fmt='RGBA', color=0xff00ffff),
])
def color(request):
    yield request.param


def test_create_background(color):
    img = Image.new(80, 60, color['fmt'], color['color'])
    px = img.get_pixel((0, 0))
    assert px == color['color']
    px = img.get_pixel((20, 10))
    assert px == color['color']


def test_draw_rectangle():
    red = 0xff0000
    green = 0x00ff00
    img = Image.new(80, 60, 'RGB', red)
    img.draw_rect((20, 20), (20, 20), green)
    px = img.get_pixel((10, 10))
    assert px == red
    px = img.get_pixel((25, 25))
    assert px == green



import builtins
import os
import sys
from typing import Optional, Any, Tuple, Union, Dict, BinaryIO, cast, TYPE_CHECKING

from PilLiteExt import ffi, lib # pylint: disable=no-name-in-module
if TYPE_CHECKING:
    from PilLiteExt.lib import ImageExt, ImageCompExt


__all__ = ['open', 'new', 'fromarray']

BMP, JPG, PNG = FORMATS = ('bmp', 'jpg', 'png')

EXT_FORMAT = {
    '.jpeg': JPG,
    '.jpg': JPG,
    '.png': PNG,
    '.bmp': BMP,
}

FORMAT_HANDLERS = {
    BMP: lib.image_to_bmp,
    JPG: (lambda img: lib.image_to_jpg(img, 100)),
    PNG: lib.image_to_png,
}

FORMAT_MAGIC = {
    JPG: b'\xFF\xD8\xFF',
    BMP: b'BM',
    PNG: b'\x89PNG',
}


def _guess_format(filename: str) -> Optional[str]:
    _, ext = os.path.splitext(filename)
    ext = ext.lower()
    return EXT_FORMAT.get(ext, None)


def _open_image(fp: BinaryIO) -> 'ImageExt':
    data = ffi.from_buffer('unsigned char[]', fp.read())
    img = lib.image_open(data, len(data))
    rv = ffi.gc(img, lib.image_free, img.width * img.height * img.components)
    if rv.buffer == ffi.NULL:
        raise IOError('Image open error')
    return rv


def _write_image(img: 'ImageExt', fp: BinaryIO, fmt: str) -> None:
    handler = FORMAT_HANDLERS[fmt]
    compressed = handler(img)
    data = ffi.buffer(compressed.buffer, compressed.size)
    fp.write(bytes(data))
    lib.image_compressed_free(compressed)


def _new_image(w: int, h: int, c: int) -> 'ImageExt':
    img = lib.image_new(w, h, c)
    rv = ffi.gc(img, lib.image_free, img.width * img.height * img.components)
    if rv.buffer == ffi.NULL:
        raise IOError('Image create error')
    return rv


def _new_image_raw(data: bytes, w: int, h: int, c: int) -> 'ImageExt':
    buf = ffi.from_buffer('unsigned char[]', data)
    img = lib.image_new_raw(buf, w, h, c)
    rv = ffi.gc(img, lib.image_free, img.width * img.height * img.components)
    if rv.buffer == ffi.NULL:
        raise IOError('Image create error')
    return rv


def open(fp: Union[str, BinaryIO], **_kwargs: Any) -> 'Image': # pylint: disable=redefined-builtin
    """
    Opens, reads and decodes the given image file.

    You can use a file object instead of a filename. File object must
    implement ``read`` method, and be opened in binary mode.
    """
    if isinstance(fp, str):
        fp = builtins.open(fp, 'rb')
        try:
            if not _is_supported(fp):
                raise IOError('Image open error')
            img = _open_image(fp)
        finally:
            fp.close()
    elif hasattr(fp, 'read'):
        if not _is_supported(fp):
            raise IOError('Image open error')
        img = _open_image(fp)
    else:
        raise ValueError
    image = Image()
    image.im = img
    return image


def new(w: int, h: int, fmt: str, bg: int) -> 'Image':
    """
    Creates new image by given size and format
    and fills it with bg color
    """
    if fmt not in ('RGB', 'RGBA', 'L', 'LA'):
        raise ValueError('invalid format')
    c = len(fmt)
    image = Image()
    image.im = _new_image(w, h, c)
    lib.image_draw_rect(image.im, 0, 0, w, h, bg)
    return image


def fromarray(obj: Any) -> 'Image':
    """ Creates new image from numpy array """
    if not hasattr(obj, '__array_interface__'):
        raise ValueError
    if str(obj.dtype) != 'uint8':
        raise ValueError
    shape = obj.shape
    if len(shape) == 2:
        w, h = shape
        components = 1
    elif len(shape) == 3:
        w, h, components = shape
        if components not in (1, 2, 3, 4):
            raise ValueError
    else:
        raise ValueError
    img = _new_image_raw(obj.tobytes(), w, h, components)
    image = Image()
    image.im = img
    return image


class Image:
    im: Optional['ImageExt'] = None

    def __init__(self) -> None:
        self.im = None

    @property
    def size(self) -> Tuple[int, int]:
        """ Returns the size of image, a tuple (width, height) """
        if not self.im:
            raise ValueError
        return (self.im.width, self.im.height)

    def save(self, fp: Union[str, BinaryIO], fmt: Optional[str] = None) -> None:
        """
        Saves this image under the given filename. If no format is
        specified, the format to use is determined from the filename
        extension, if possible.

        You can use a file object instead of a filename. The file object
        must implement the ``write`` method, and be opened in binary mode.
        """
        if not self.im:
            raise ValueError

        if isinstance(fp, str):
            filename = fp
            fp = builtins.open(filename, 'wb')
            try:
                if not fmt:
                    fmt = _guess_format(filename)
                if fmt not in FORMATS:
                    raise ValueError(f'unsupported format {fmt}')
                _write_image(self.im, fp, fmt)
            finally:
                fp.close()
        elif hasattr(fp, 'write'):
            if not fmt and hasattr(fp, 'name'):
                filename = getattr(fp, 'name')
                fmt = _guess_format(filename)
            if fmt not in FORMATS:
                raise ValueError(f'unsupported format {fmt}')
            _write_image(self.im, fp, fmt)
        else:
            raise ValueError

    def resize(self, size: Tuple[int, int]) -> 'Image':
        """ Returns a resized copy of this image. """
        w, h = size
        if not w or not h:
            raise ValueError('invalid size')
        if not self.im:
            raise ValueError
        image = Image()
        resized = lib.image_resize(self.im, w, h)
        image.im = ffi.gc(resized, lib.image_free, w * h * resized.components)
        return image

    def thumbnail(self, size: Tuple[int, int]) -> None:
        """
        Make this image into a thumbnail. This method modifies the
        image to contain a thumbnail version of itself, no larger than
        the given size and preserving original aspect ratio.
        """
        if not self.im:
            raise ValueError
        w, h = size
        x, y = self.size
        if x > w:
            ratio = y / x
            x, y = w, int(ratio * w)
        if y > h:
            ratio = x / y
            x, y = int(ratio * h), h
        resized = self.resize((x, y))
        self.im = resized.im

    def show(self) -> None:
        if not self.im:
            raise ValueError

        from tempfile import NamedTemporaryFile
        from subprocess import run
        with NamedTemporaryFile(suffix='.png') as fp:
            fp = cast(BinaryIO, fp)
            _write_image(self.im, fp, PNG)
            fp.flush()
            if sys.platform == 'linux':
                run(['display', fp.name])

    def put_pixel(self, coord: Tuple[int, int], color: int) -> None:
        """ Draws pixel at (x, y) coord """
        if not self.im:
            raise ValueError
        x, y = coord
        lib.image_put_pixel(self.im, x, y, color)

    def get_pixel(self, coord: Tuple[int, int]) -> int:
        """ Get pixel at (x, y) coord """
        if not self.im:
            raise ValueError
        x, y = coord
        return lib.image_get_pixel(self.im, x, y)

    def draw_rect(self, coord: Tuple[int, int], size: Tuple[int, int], color: int) -> None:
        """ Draws rectagnle at (x, y) coord with (w, h) size"""
        if not self.im:
            raise ValueError
        x, y = coord
        w, h = size
        lib.image_draw_rect(self.im, x, y, w, h, color)

    def tobytes(self) -> bytes:
        """ Raw bytes of image data """
        if not self.im:
            raise ValueError
        size = self.im.width * self.im.height * self.im.components
        return ffi.buffer(self.im.buffer, size)

    @property
    def __array_interface__(self) -> Dict:
        """ Numpy array support """
        if not self.im:
            raise ValueError
        w, h, components = self.im.width, self.im.height, self.im.components
        return {
            'version': 3,
            'shape': (w, h, components),
            'typestr': '|u1',
            'data': self.tobytes(),
        }


def _get_magic_mime(fp: BinaryIO, n: int = 4) -> Optional[bytes]:
    try:
        pos = fp.tell()
        data = fp.read(n)
        if not data or len(data) != n:
            return None
        return data
    except EOFError:
        return None
    finally:
        fp.seek(pos)


def _is_supported(fp: BinaryIO) -> bool:
    buff = _get_magic_mime(fp)
    if not buff:
        return False
    return any(buff.startswith(magic) for magic in FORMAT_MAGIC.values())



import os
from io import BytesIO
from tempfile import TemporaryDirectory
import pytest
from PilLite import Image

BASE_DIR = os.path.dirname(__file__)
SIZE = (256, 184)


@pytest.fixture(scope='function')
def tmpdir():
    with TemporaryDirectory() as folder:
        yield folder


@pytest.fixture(params=[
    '.png', '.jpg', '.bmp'
], scope='function')
def infile(request):
    ext = request.param
    yield os.path.join(BASE_DIR, 'data', 'image' + ext)


@pytest.fixture(params=[
    '.png', '.jpg', '.bmp'
], scope='function')
def outfile(tmpdir, request):
    ext = request.param
    yield os.path.join(tmpdir, 'image' + ext)


@pytest.fixture
def png_file():
    yield os.path.join(BASE_DIR, 'data', 'image.png')


@pytest.fixture
def missing_file():
    yield os.path.join(BASE_DIR, 'data', 'missing.png')


@pytest.fixture(params=[
    '.tif', '.tga'
])
def unsupported_file(request):
    ext = request.param
    yield os.path.join(BASE_DIR, 'data', 'image' + ext)


def test_open(infile):
    img = Image.open(infile)
    assert img.size == SIZE


def test_open_file_desciptor(infile):
    with open(infile, 'rb') as fd:
        img = Image.open(fd)
        assert img.size == SIZE


def test_open_and_save(infile, outfile):
    img = Image.open(infile)
    img.save(outfile)
    assert os.path.exists(outfile)
    img = Image.open(outfile)
    assert img.size == SIZE


def test_open_and_save_file_descriptor(infile, outfile):
    img = Image.open(infile)
    with open(outfile, 'wb') as fd:
        img.save(fd)
    assert os.path.exists(outfile)
    img = Image.open(outfile)
    assert img.size == SIZE


def test_bytes_io(png_file):
    with open(png_file, 'rb') as fd:
        data = BytesIO(fd.read())
    img = Image.open(data)
    assert img.size == SIZE


@pytest.fixture(params=[(128, 90), (512, 360)])
def resize(request):
    yield request.param


def test_resize(png_file, resize):
    img = Image.open(png_file)
    img2 = img.resize(resize)
    assert img2 is not img
    assert img2.size == resize
    assert img.size == SIZE


def test_thumbnail(png_file):
    img = Image.open(png_file)
    img.thumbnail((128, 90))
    assert img.size == (125, 90)


def test_not_found(missing_file):
    with pytest.raises(IOError):
        Image.open(missing_file)


def test_unsupported(unsupported_file):
    with pytest.raises(IOError):
        Image.open(unsupported_file)


def test_save_unsupported(png_file, tmpdir):
    img = Image.open(png_file)
    outfile = os.path.join(tmpdir, 'image.tif')
    with pytest.raises(ValueError):
        img.save(outfile)


def test_save_unsupported2(png_file, tmpdir):
    img = Image.open(png_file)
    outfile = os.path.join(tmpdir, 'image')
    with pytest.raises(ValueError):
        img.save(outfile, 'tiff')


@pytest.fixture(params=[
    b'\xFF\xD8\xFF\xD9',
    b'BM\xFF\xFF',
    b'\x89PNG\xFF\xFF',
    b'\xA0\xB0',
])
def broken_data(request):
    yield request.param


def test_corrupt_png_failure_string(broken_data):
    with pytest.raises(IOError, match='Image open error'):
        data = BytesIO(broken_data)
        Image.open(data)


@pytest.fixture(params=[(0, 100), (100, 0)])
def broken_size(request):
    yield request.param


def test_invalid_size_resize(png_file, broken_size):
    with pytest.raises(ValueError, match='invalid size'):
        img = Image.open(png_file)
        img.resize(broken_size)


def test_invalid_size_thumbnail(png_file, broken_size):
    with pytest.raises(ValueError, match='invalid size'):
        img = Image.open(png_file)
        img.thumbnail(broken_size)



import os
from cffi import FFI

THIS_DIR = os.path.dirname(__file__)
STB_DIR = os.path.join(THIS_DIR, 'stb')


with open(os.path.join(THIS_DIR, 'lib.h'), 'r') as f:
    ctext = f.read()

ffibuilder = FFI()

ffibuilder.cdef(ctext)

ffibuilder.set_source("PilLiteExt", """
    #include "lib.h"
""",
    sources=[
        os.path.join(THIS_DIR, 'lib.c'),
        os.path.join(STB_DIR, 'stb_image.c'),
        os.path.join(STB_DIR, 'stb_image_resize2.c'),
        os.path.join(STB_DIR, 'stb_image_write.c'),
    ],
    include_dirs=[THIS_DIR, STB_DIR],
    define_macros=[
        ('STBI_NO_PSD', None),
        ('STBI_NO_TGA', None),
        ('STBI_NO_GIF', None),
        ('STBI_NO_HDR', None),
        ('STBI_NO_PIC', None),
        ('STBI_NO_PNM', None),
    ])

if __name__ == "__main__":
    ffibuilder.compile()



__all__ = [
    '__title__', '__summary__', '__uri__', '__version__', '__author__',
    '__email__', '__license__', '__copyright__',
]

__title__ = 'Pil-Lite'
__summary__ = 'Python Imaging Library Lite'
__uri__ = 'https://github.com/alexa-infra/pil-lite'

__version__ = '0.1.3'

__author__ = 'Alexey Vasilyev'
__email__ = 'alexey@alexadotlife.com'

__license__ = 'MIT'
__copyright__ = 'Copyright 2014-2024 {0}'.format(__author__)



import pytest

try:
    import numpy as np
except ImportError:
    pytest.skip('Numpy is not installed', allow_module_level=True)

from PilLite import Image


def test_to_array():
    img = Image.open('test/data/image.png')
    im = np.array(img)
    assert im.ndim == 3
    assert im.dtype == np.dtype('uint8')


def test_to_array_color():
    img = Image.new(10, 10, 'RGB', 0xff0000)
    im = np.array(img)
    assert im.ndim == 3
    assert im.dtype == np.dtype('uint8')
    pixel = np.array([255, 0, 0], 'u1')
    np.testing.assert_array_equal(im[0, 0], pixel)
    arr = np.tile(pixel, 10 * 10).reshape((10, 10, 3))
    np.testing.assert_array_equal(im, arr)


def test_from_array():
    pixel = np.array([255, 0, 0], 'u1')
    arr = np.tile(pixel, 10 * 10).reshape((10, 10, 3))
    img = Image.fromarray(arr)
    assert img.size == (10, 10)
    arr_back = np.array(img)
    np.testing.assert_array_equal(arr, arr_back)
